// SkypeMorph project
// 
// SMClientMain.cpp
// 
#include "UProxy.h"
#include "Morpher.h"
#include "SkypeClient.h"
#include "Util.h"
#include "version.h"
#include "debug.h"
#include <cassert>
#include <iostream>
#include <string>

#include "config.h"

using namespace boost::asio;

namespace {

typedef struct {
	std::string host;
	std::string addr;
	int port;
	int udp_port;
	int skype_port;
	int method;
	std::string size_dist_fn;
	std::string time_dist_fn;
	std::string tor_size_dist_fn;
	std::string matrix_fn;
	std::string skyperc;
	std::string srv_account;
	std::string logfile;
} Arguments;

void relayError(const boost::system::error_code &err) {
	std::cerr << "Error with relay, exiting..." << std::endl;
	exit(0);
}


void printUsage(char *name) {
	std::cerr << "Usage: " << name << " <options> " << std::endl << std::endl
			<< "Options:" << std::endl
			<< "-i <if / addr>      -- Specify the interface or the ip address of the SOCKS." << std::endl
			<< "-I <addr>           -- Specify the external ip address of the client." << std::endl
			<< "-p <local port>     -- Local port for SOCKS proxy interface." << std::endl
			<< "-u <local udp port> -- Local port for UDP interface." << std::endl
			<< "-q <skype port>     -- Local port for Skype interface." << std::endl
			<< "-m <method>         -- Morphing method to use. Possible values are " << std::endl
			<< "   0                => fixed size and timing" << std::endl
			<< "   1                => matrix-based morphing" << std::endl
			<< "   2                => naive morphing"        << std::endl
			<< "-k <skyperc>        -- Skype configure file" << std::endl
			<< "-a <srv account>    -- Skype account for the server" << std::endl
			<< "-l <log file>       -- log filename" << std::endl
			<< "-h                  -- Print this help message." << std::endl << std::endl;

	exit(0);
}

int parseOptions(int argc, char **argv, Arguments *arguments) {
	int c;

	arguments->host = "127.0.0.1";
	arguments->addr = "0.0.0.0";
	arguments->port = 5060;
	arguments->udp_port = 0;      // default: no specific udp port
	arguments->skype_port = 30000;
	arguments->method = 0;        // default method
	arguments->size_dist_fn = "";
	arguments->time_dist_fn = "";
	arguments->tor_size_dist_fn = "";
	arguments->matrix_fn   = "";
	arguments->skyperc = "skyperc";
	arguments->srv_account = "";
	arguments->logfile = "client.log";

	opterr = 0;

	while ((c = getopt (argc, argv, "a:i:I:p:u:q:s:t:S:m:x:k:l:h")) != -1) {
		switch (c) {
		case 'i':
			arguments->host = optarg;

			break;
		case 'I':
			arguments->addr = optarg;
			break;
		case 'p':
			arguments->port = atoi(optarg);
			break;
		case 'u':
			arguments->udp_port = atoi(optarg);
			break;
		case 'q':
			arguments->skype_port = atoi(optarg);
			break;
		case 's':
			arguments->size_dist_fn = optarg;
			break;
		case 't':
			arguments->time_dist_fn = optarg;
			break;
		case 'S':
			arguments->tor_size_dist_fn = optarg;
			break;
		case 'm':
			arguments->method = atoi(optarg);
			break;
		case 'x':
			arguments->matrix_fn = optarg;
			break;
		case 'k':
			arguments->skyperc = optarg;
			break;
		case 'a':
			arguments->srv_account = optarg;
			break;
		case 'l':
			arguments->logfile = optarg;
			break;
		case 'h':
			printUsage(argv[0]);
		default:
			std::cerr << "Unknown option: " << c << std::endl;
			printUsage(argv[0]);
		}
	}


	if (arguments->method != 0 &&
			(arguments->size_dist_fn == "" ||
					arguments->time_dist_fn == "" ||
					arguments->matrix_fn   == "")) {
		std::cerr << "non default morphing method given, and missing "
				<< "distribution data" << std::endl;
		return 0;
	}

	return 1;
}

} // anonymous namespace


int main(int argc, char** argv) {
	Arguments arguments;


	std::cout << "skypemorph version r" << skypemorph_revision << std::endl;

	if (!parseOptions(argc, argv, &arguments)) {
		printUsage(argv[0]);

		return 2;
	}

	char* k;
	char* v;
	char* result;

	Config cf;
	cf.load_cfg_file(arguments.skyperc.c_str());

	if (!log_init(arguments.logfile.c_str())) {
		std::cerr << "[main] " << "cannot open " << arguments.logfile
				<< ", exit now..." << std::endl;
		exit(-1);
	}



	std::cout << "Running SkypeKit Runtime" << std::endl;
	char* command = (char*) malloc(strlen((char*)cf.options["skypekitadr"])+strlen((char* )cf.options["skrtport"])+9);
	sprintf (command, "%s %s %s&",(char*)(cf.options["skypekitadr"]),"-m -p",(char* )cf.options["skrtport"]);
	system(command);

	Morpher * morpher = new Morpher(atoi(cf.options["morphingmethod"]));
	if (cf.options["morphingmethod"] != 0) {
		morpher->init(cf.options["skypetimecdf"],
				cf.options["skypesizecdf"],
				cf.options["torsizecdf"],
				cf.options["morphingmatrix"]);
	}


	SkypeClient * skypeclient = new SkypeClient(cf.options["ipadr"],
			atoi(cf.options["udpport"]),
			//atoi(cf.options["skypeport"]),
			cf,
			cf.options["serverid"]);

	std::cerr << "[main] " << "proxy starting.." << std::endl;

	boost::asio::io_service io_service;
	UDPRelay *udprelay = new UDPRelay(io_service, atoi(cf.options["udpport"]),
			boost::bind(relayError,
					placeholders::error));


	Proxy *proxy = new Proxy(udprelay, io_service, atoi(cf.options["socksport"]), morpher, skypeclient);

	boost::shared_ptr<Codec> codec(new Codec(false));
	std::string host;
	uint16_t port;
	skypeclient->runOnce(&host, &port, codec);

	proxy->relay->openStreamWOSocks(host, port,
			codec, proxy->_morpher);
	proxy->relay->_pkzer->run();

	std::cerr << "[main] " << "SOCKS proxy ready on " << atoi(cf.options["socksport"]) << "." << std::endl;


	io_service::work work(io_service);
	io_service.run();
}

