// SkypeMorph project
// 
// SMServerMain.cpp
// 
#include "Server.h"
#include "Util.h"
#include "SkypeServer.h"
#include "version.h"
#include "debug.h"
#include "config.h"

#include <boost/thread.hpp>

using namespace boost::asio;

namespace {

typedef struct {
	std::string addr;
	int port;
	int remote_port;
	int skype_port;
	int method;
	std::string size_dist_fn;
	std::string time_dist_fn;
	std::string tor_size_dist_fn;
	std::string matrix_fn;
	std::string skyperc;
	std::string srv_account;
	std::string logfile;
} Arguments;

void printUsage(char *name) {
	std::cerr << "Usage: " << name << " <options> " << std::endl << std::endl
			<< "Options:" << std::endl
			<< "-i <if / addr>      -- Specify the interface or the ip address of the UDP interface." << std::endl
			<< "-I <addr>           -- Specify the external ip address of the client." << std::endl
			<< "-p <local udp port> -- Local port for UDP interface." << std::endl
			<< "-q <skype port>     -- Local port for Skype interface." << std::endl
			<< "-r <local tcp port> -- Local port for TCP relay target." << std::endl
			<< "-m <method>         -- Morphing method to use. Possible values are " << std::endl
			<< "   0                => fixed size and timing" << std::endl
			<< "   1                => matrix-based morphing" << std::endl
			<< "   2                => naive morphing"        << std::endl
			<< "-k <skyperc>        -- Skype configure file" << std::endl
			<< "-a <srv account>    -- Skype account for the server" << std::endl
			<< "-l <log file>       -- log filename" << std::endl
			<< "-h                  -- Print this help message." << std::endl << std::endl;

	exit(0);
}

int parseOptions(int argc, char **argv, Arguments *arguments) {
	int c;

	arguments->addr = "0.0.0.0";
	arguments->port = 20000;
	arguments->remote_port = 9000;
	arguments->skype_port = 30000;
	arguments->method = 0;        // default method
	arguments->size_dist_fn = "";
	arguments->time_dist_fn = "";
	arguments->tor_size_dist_fn = "";
	arguments->matrix_fn   = "";
	arguments->skyperc = "skyperc";
	arguments->srv_account = "";
	arguments->logfile = "server.log";

	opterr = 0;

	while ((c = getopt (argc, argv, "I:p:r:q:s:t:S:m:x:k:a:l:h")) != -1) {
		switch (c) {
		case 'I':
			arguments->addr = optarg;
			break;
		case 'p':
			arguments->port = atoi(optarg);
			break;
		case 'r':
			arguments->remote_port = atoi(optarg);
			break;
		case 'q':
			arguments->skype_port = atoi(optarg);
			break;

		case 's':
			arguments->size_dist_fn = optarg;
			break;
		case 't':
			arguments->time_dist_fn = optarg;
			break;
		case 'S':
			arguments->tor_size_dist_fn = optarg;
			break;
		case 'm':
			arguments->method = atoi(optarg);
			break;
		case 'x':
			arguments->matrix_fn = optarg;
			break;
		case 'k':
			arguments->skyperc = optarg;
			break;
		case 'a':
			arguments->srv_account = optarg;
			break;
		case 'l':
			arguments->logfile = optarg;
			break;
		case 'h':
			printUsage(argv[0]);
		default:
			std::cerr << "Unknown option: " << c << std::endl;
			printUsage(argv[0]);
		}
	}

	if (arguments->method != 0 &&
			(arguments->size_dist_fn == "" ||
					arguments->time_dist_fn == "" ||
					arguments->matrix_fn   == "")) {
		std::cerr << "non default morphing method given, and missing "
				<< "distribution data" << std::endl;
		return 0;
	}

	return 1;
}

} // anonymous namespace 

int main(int argc, char** argv) {
	Arguments arguments;
	int ret = 0;

	std::cout << "skypemorph version r" << skypemorph_revision << std::endl;

	if (!parseOptions(argc, argv, &arguments)) {
		printUsage(argv[0]);
		ret = 2;
		return ret;
	}


	Config cf;
	cf.load_cfg_file(arguments.skyperc.c_str());

	if (!log_init(arguments.logfile.c_str())) {
		std::cerr << "[main] " << "cannot open " << arguments.logfile
				<< ", exit now..." << std::endl;
		exit(-1);
	}


	std::cout << "Running SkypeKit Runtime" << std::endl;
	char* command = (char*) malloc(strlen((char*)cf.options["skypekitadr"])+strlen((char* )cf.options["skrtport"])+9);
	sprintf (command, "%s %s %s&",(char*)(cf.options["skypekitadr"]),"-m -p",(char* )cf.options["skrtport"]);
	system(command);

	Morpher * morpher = new Morpher(atoi(cf.options["morphingmethod"] ));
	if (cf.options["morphingmethod"] != 0) {
		morpher->init(cf.options["skypetimecdf"],
				cf.options["skypesizecdf"],
				cf.options["torsizecdf"],
				cf.options["morphingmatrix"]);
	}

	std::cerr << "[main] " << "server starting.." << std::endl;

	std::cout<<cf.options["remoteport"]<<std::endl;
	Server * server = new Server(atoi(cf.options["udpport"]), atoi(cf.options["remoteport"]), morpher);

	SkypeServer * skypeserver = new SkypeServer(cf.options["ipadr"], atoi(cf.options["udpport"]), //atoi(cf.options["skypeport"])
			cf,
			server);

	try {
		boost::thread & tskype  = skypeserver->run();
		boost::thread & tserver = server->run();

		tserver.join();
		tskype.join();
	}
	catch (...) {
		std::cerr << "[main] " << "caught exception" << std::endl;
		server->stop();

		skypeserver->stop();
		ret = -1;

		goto cleanup;
	}

	std::cerr << "[main] " << "finished" << std::endl;


	cleanup:
	delete server;
	delete morpher;

	done:
	return ret;
}

