#include "Server.h"

#include <boost/shared_ptr.hpp>
#include <boost/asio.hpp>
#include <boost/bind.hpp>
#include <cassert>
#include "Packetizer.h"
#include "Relay.h"
#include "RelayStream.h"
#include "ProxyShuffler.h"
#include "Crypto.h"
#include "Util.h"
#include "debug.h"

using namespace boost::asio;

Server::Server(int listenPort, uint16_t remotePort, Morpher * morpher) 
  : _rport(remotePort), _morpher(morpher) 
{

  ip::udp::socket * socket = new ip::udp::socket(_io_service, 
                                   ip::udp::endpoint(ip::udp::v4(), listenPort)); 
  _udpsocket = boost::shared_ptr<ip::udp::socket>(socket); 
}

void Server::serverSession() {
  std::cerr << "[server] " << "ready on " << _udpsocket->local_endpoint() << std::endl;
  recvUDPData();
  boost::asio::io_service::work work(_io_service);
  _io_service.run(); 
}

boost::thread & Server::run() {
  _thread = boost::thread(&Server::serverSession, this);
  return _thread; 
}

void Server::stop() {
  _io_service.stop(); 

  _thread.join(); 
}


void Server::recvUDPData() {
  _udpsocket->async_receive_from(buffer(_buf, sizeof(_buf)), 
                                 _sender_ep, 
                                 boost::bind(&Server::handleUDPData, 
                                             this, placeholders::error, 
                                             placeholders::bytes_transferred)); 
}

void Server::handleUDPData(const boost::system::error_code &err, 
                           std::size_t bytes_transferred) 
{
  if (err) {
    std::cerr << "[server] " << "Error accepting incoming connection: " << err << std::endl;
    recvUDPData();

    return;
  }

#ifdef DEBUG_UDP_DISPATCH
  std::cerr << "[server] " << "Got UDP " << bytes_transferred << " bytes " 
            << "from " << _sender_ep << std::endl; 
#endif

  if (_clients.count(_sender_ep) > 0) {
    boost::shared_ptr<Client> c = _clients[_sender_ep];
    c->push_recv_msg(_buf, bytes_transferred);
  } 
  else {
    std::cerr << "[server] " << "Ignore unknown client " << _sender_ep << std::endl; 
  }

  recvUDPData();
}

bool Server::verifyConnectionRequest(boost::shared_ptr<Client> & c, 
                                     boost::shared_ptr<Packetizer> & pkzer) {
  UDPPacket * pkt = c->pop_recv_msg(); 
  if (pkzer->verifyHello(pkt)) {
    delete pkt;

    pkt = pkzer->genHello(); 
    _udpsocket->send_to(buffer(pkt->data, pkt->length), c->endpoint()); 
    delete pkt; 

    std::cerr << "[server] " << "UDP connection established with " << c->endpoint() << std::endl;

    return true; 
  }

  return false;
}

void relayError(const boost::system::error_code &err) {
  std::cerr << "[server] " << "Error with relay, exiting..." << std::endl;
  exit(0);
}

void Server::clientSession(boost::shared_ptr<Client> & c, 
                           boost::shared_ptr<Codec> & codec) {
  std::cerr << "[server] " << "Client " << c->endpoint() 
            << " starts..." << std::endl; 
  boost::asio::io_service io_service;

  boost::shared_ptr<Packetizer> pkzer(new Packetizer(c, codec, _morpher)); 

  // if (!verifyConnectionRequest(c, pkzer)) {

  //   _clients.erase(c->endpoint()); 

  //   std::cerr << "[server] " << " invalid connection request. exit now" << std::endl
  //             << "Client " << c->endpoint() << " stops..." << std::endl;
  //   return; 

  // }

  std::string tor_addr = "127.0.0.1"; 
  uint16_t    tor_port = _rport; 

  boost::shared_ptr<Relay> relay(new Relay(io_service, 
                                           boost::bind(relayError, 
                                                       placeholders::error))); 

  try {
    boost::shared_ptr<RelayStream> rstream(relay->openStream(tor_addr, tor_port));
    boost::asio::io_service::work * work = new boost::asio::io_service::work(io_service);
    boost::shared_ptr<ProxyShuffler> proxyShuffler(new ProxyShuffler(rstream, pkzer, work)); 

    pkzer->run(); 
    proxyShuffler->shuffle();

    io_service.run(); 
  } 
  catch (boost::system::system_error & error) {
    std::cerr << "[server] " << " openStream filed: " << error.what() << std::endl 
              << " eixt now." << std::endl; 
  }
  catch (...) {
    std::cerr << "[server] " << " unexpected error. exit now" << std::endl; 
  }


  _clients.erase(c->endpoint()); 

  std::cerr << "[server] " << "Client " << c->endpoint() 
            << " stops..." << std::endl << std::endl; 
}


void Server::newClientSession(const std::string & addr, uint16_t port, boost::shared_ptr<Codec> & codec) {
  ip::udp::endpoint ep(ip::address::from_string(addr.c_str()), port); 
  if (_clients.count(ep) != 0) {
    std::cerr << "[server] " << "session exists for " << ep << std::endl; 

    return;
  }
  boost::shared_ptr<Client> c(new Client(_udpsocket, ep));
  std::cerr << "[server] " << "new client " << ep << std::endl; 
  _clients[ep] = c;

  c->thread = boost::thread(&Server::clientSession, this, c, codec);
}


